#!/usr/bin/env python


import os, sys, gtk, random

############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time

import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb

import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)
            #self.set_single_line_mode(False)

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message

#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2012 Brian Hahn, 2024 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP. It builds on an original script by Brian Hahn from 2011.
#
# It can be executed by selecting the menu option: 'Filters>Path Copy Rotate Shift Wander...

# Initial Version
#	Options for moving and rotating one path along another.
#	Options for stroking the paths in a variety of colours
#	Options for dealing with the generated paths
# Version 0.01
#	Added scaling and scaling positioning compensation
# Version 0.02
#	Made 1 repetition available from UI
#   Detect open/closed copy paths and adjust position of repeats accordingly
#	Option for stroking path with brush
#	Option to choose to use scaling positioning compensation
# Version 0.03
#	Fixed zero walk length path bug
# Version 0.04
#	Added half, quarter turns to force revolution option
#	Added option to set value for angle of first copy path of each repeat
#   Extended colour options to reverse gradients/ alternately applied gradients
#   Corrected BG-FG alternate path colouring
# Version 0.05
#    Not uploaded/published
#    Restored older angle modifier values
# Version 0.06 Tabbed
#    Added tabs
#    Altered dialog order/wording

#from gimpfu import *
#from array import array
#import math
#import random

#from gtk import gdk
#w = gdk.screen_width()
#h = gdk.screen_height()

# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)
#darkgrey = gimpcolor.RGB(50, 50, 50)
#midgrey = gimpcolor.RGB(125, 125, 125)
#lightgrey = gimpcolor.RGB(170, 170, 170)
#verylightgrey = gimpcolor.RGB(222, 222, 222)
#red = gimpcolor.RGB(220,10,10)	



#### Main Filter Proc

#def do_stuff(image, layer, pf_offset_x, pf_offset_y, pf_angle, initial_angle, force_full_turns, pf_offset_x_mod, pf_offset_y_mod, pf_angle_mod, num_travel_steps, scale_target, scale_compensation, num_copies, path_width, stroke_method, modifier, modifier_options, pattern, colour_option, path_options, ptab):


def do_stuff(image, layer,  path_width, stroke_method, colour_option, num_copies, pf_offset_x, pf_offset_y, 
            initial_angle, pf_angle, force_full_turns, num_travel_steps, 
            scale_target, scale_compensation, pf_offset_x_mod, pf_offset_y_mod, pf_angle_mod, modifier_options, modifier, 
            pattern, path_options, ptab):
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)
	
    # temp values
    use_travel_path = True
    travel_step_size = 100
	
    
    layer_list = image.layers
    num_paths, vector_ids = pdb.gimp_image_get_vectors(image)
    vector_list = image.vectors
	
    if "walk" in str(vector_list) and num_paths>1 and use_travel_path: 
	
        if force_full_turns>0:
            if abs(pf_angle)>=1.0:
                if force_full_turns == 1:
                    num_copies = int(360.0/abs(pf_angle))
                elif force_full_turns == 2:
                    num_copies = int(180.0/abs(pf_angle))+1
                elif force_full_turns == 3:
                    num_copies = int(90.0/abs(pf_angle))+1
                elif force_full_turns == 4:
                    num_copies = int(270.0/abs(pf_angle))+1
            else:
                if abs(pf_angle)<0:
                    pf_angle = -1.0
                else:
                   pf_angle = 1.0
                if force_full_turns == 1:
                    num_copies = 360
                elif force_full_turns == 2:
                    num_copies = 180
                elif force_full_turns == 3:
                    num_copies = 90
                elif force_full_turns == 4:
                    num_copies = 270
				
	
        travel_vector = pdb.gimp_image_get_vectors_by_name(image, "walk")
        original_fg_colour = pdb.gimp_context_get_foreground()
        original_bg_colour = pdb.gimp_context_get_background()
		
        # Get travel path length
        NumberOfStrokes_in_Path,ActivePath_StrokeIDs_List = pdb.gimp_vectors_get_strokes(travel_vector)
        ActivePath_StrokeID = ActivePath_StrokeIDs_List[0]
        travel_vector_length = pdb.gimp_vectors_stroke_get_length(travel_vector,ActivePath_StrokeID,0.01)
        # pdb.gimp_message("travel_vector_length = " + str(travel_vector_length))

        # find out if travel path is closed or open
        open_path = False
        NumberOfStrokes_in_TravelPath,TravelPath_StrokeIDs_List = pdb.gimp_vectors_get_strokes(travel_vector)
        type, num_points, controlpoints, closed = pdb.gimp_vectors_stroke_get_points(travel_vector, TravelPath_StrokeIDs_List[0])
        if not closed: open_path = True
		
        # use these coordinates when travel vector has a length of zero
        first_control_x, first_control_y = controlpoints[2], controlpoints[3]

        # determine position of work vector path in vector stack - ie the path which will be moved and rotated along the travel path
        if vector_list[0] == travel_vector:
            work_vector = vector_list[1]
        else:
            work_vector = vector_list[0]
			
        # work out centre of work path and offsets to move a copy of the path to the image origin before starting
        NumberOfStrokes_in_WorkPath,WorkPath_StrokeIDs_List = pdb.gimp_vectors_get_strokes(work_vector)
        type, num_points, controlpoints, closed = pdb.gimp_vectors_stroke_get_points(work_vector, WorkPath_StrokeIDs_List[0])
        stroke_loop = 0
        maxX,minX,maxY,minY, centreX, centreY = fn_get_min_max_path(work_vector)
        work_vector_x_offset = 0-(minX + centreX)
        work_vector_y_offset = 0-(minY + centreY)
		
        # copy and add the work vector path to the image
        copy_vector = pdb.gimp_vectors_copy(work_vector)
        pdb.gimp_image_insert_vectors(image, copy_vector, None, -1)
        # move the new path to the image origin
        while stroke_loop<NumberOfStrokes_in_WorkPath:
            Char_Vector_Stroke = WorkPath_StrokeIDs_List[stroke_loop]
            pdb.gimp_vectors_stroke_scale(copy_vector, Char_Vector_Stroke, 1, 1)#######
            pdb.gimp_vectors_stroke_translate(copy_vector, Char_Vector_Stroke, work_vector_x_offset, work_vector_y_offset)
            stroke_loop += 1
        
        # work out size of steps along travel path
        if travel_vector_length==0:
            travel_step_size = 0
            num_scale_steps=1
            num_travel_steps = 1
        else:
            if open_path:
                if num_travel_steps>1:
                    travel_step_size = travel_vector_length/((num_travel_steps-1))
                    num_scale_steps=num_travel_steps-1
                else:
                    travel_step_size = travel_vector_length/2
                    num_scale_steps=1
            else:
                travel_step_size = travel_vector_length/(num_travel_steps)
                num_scale_steps=num_travel_steps

        if travel_step_size<0: travel_step_size = int(travel_vector_length/2)
        get_out_now = False
        travel_loop = 0
		
        scaling=1
        if num_travel_steps == 1: scaling += scale_target
		
		
        scale_step = (scale_target)/num_travel_steps
        comp_value = 1.00
        compensator = [1.00]
        comp_loop = 0
        # create a list to hold scaling steps
        while comp_loop<num_scale_steps:
            comp_value += scale_step
            compensator.append(comp_value)
            comp_loop += 1
            
        compensator.reverse()
        # scale the copy path
        while travel_loop<num_travel_steps:
            NumberOfStrokes_in_WorkPath,WorkPath_StrokeIDs_List = pdb.gimp_vectors_get_strokes(work_vector)
            type, num_points, controlpoints, closed = pdb.gimp_vectors_stroke_get_points(work_vector, WorkPath_StrokeIDs_List[0])
            stroke_loop = 0
            while stroke_loop<NumberOfStrokes_in_WorkPath:
                Char_Vector_Stroke = WorkPath_StrokeIDs_List[stroke_loop]
                pdb.gimp_vectors_stroke_scale(copy_vector, Char_Vector_Stroke, scaling, scaling)#######
                stroke_loop += 1
				
            if scaling + scale_step>(0.00 + abs(scale_step) ):
                scaling += scale_step
            else:
                get_out_now = True
		
            if travel_vector_length == 0:
                dist=0
                x_point = first_control_x
                y_point = first_control_y
            else:
                if num_travel_steps == 1:
                    dist = int(travel_vector_length/2)
                else:
                    if scale_compensation == 1:
                        dist = (travel_loop * travel_step_size) / compensator[travel_loop] # attempt to add some scaling to the distances between repeats
                    else:
                        dist = (travel_loop * travel_step_size)
				
                if dist>travel_vector_length: dist = travel_vector_length
                x_point,y_point,SlopeAtThisPoint,IsThisPointValid = pdb.gimp_vectors_stroke_get_point_at_dist(travel_vector,ActivePath_StrokeID,dist,0.01)

            pdb.gimp_item_set_visible(work_vector, False)
			
            active_layer = pdb.gimp_image_get_active_layer(image)
            if active_layer == -1: active_layer = pdb.gimp_image_set_active_layer(image, layer_list[0])
			
            pdb.gimp_context_set_stroke_method(0) # Use line with stroke method
			
            # get current colours
            pdb.gimp_context_set_foreground(original_fg_colour)
            pdb.gimp_context_set_background(original_bg_colour)
            original_current_gradient = pdb.gimp_context_get_gradient()
            current_gradient = original_current_gradient
			
            # get gradient colour information
            list_of_gradient_positions = fn_get_gradient_positions(num_copies)
            if colour_option == 7:
                reverse_gradient = True
            else:
                reverse_gradient = False
            number_of_samples, colours_of_samples = pdb.gimp_gradient_get_custom_samples(current_gradient, len(list_of_gradient_positions), list_of_gradient_positions,reverse_gradient)
			
            if colour_option == 8:
                number_of_samples, colours_of_reversed_samples = pdb.gimp_gradient_get_custom_samples(current_gradient, len(list_of_gradient_positions), list_of_gradient_positions,True)
                number_of_samples, colours_of_samples = pdb.gimp_gradient_get_custom_samples(current_gradient, len(list_of_gradient_positions), list_of_gradient_positions,False)
			
            pf_centre_X = x_point+work_vector_x_offset
            pf_centre_Y = y_point+work_vector_y_offset
			
			
            # create modifier pattern list according to ui pattern setting
            pattern_list = [0.0]
            pattern_max = abs(modifier)
		
            if pattern == 0: # ripple type
                positive = True
                wloop = 1.0
                while wloop>=(0.0 - modifier):
                    if positive:
                        pattern_list.append(wloop)
                        wloop+=1.0
                        if wloop>pattern_max: 
                            positive=False
                            wloop = pattern_max-1
                
                    else:
                        pattern_list.append(wloop)
                        wloop-=1
				
                wloop = 0 - abs(pattern_max-1)
                pattern_bit_list = [wloop]
                while wloop<0:
                    wloop+=1
                    pattern_bit_list.append(wloop)
			
			
                pattern_list.extend(pattern_bit_list)
                pattern_list.pop()
			
            elif pattern ==1: # ascending type
                wloop = 1.0
                while wloop<=(modifier):
                    pattern_list.append(wloop)
                    wloop+=1.0
		
            elif pattern ==2: # descending type
                wloop = 1.0
                while wloop<=(modifier):
                    pattern_list.append(wloop)
                    wloop+=1.0
		
                pattern_list.reverse()
			
            elif pattern ==3: # large zigzag type
                pattern_list.append(modifier)
                pattern_list.append(0.0)
                pattern_list.append(0.0 - modifier)
			
            elif pattern ==4: # small zigzag type
                pattern_list.append(modifier)
				
					
            loop = 0 # was 1
            #while loop<=num_copies:
            while loop<num_copies:
                new_vector = pdb.gimp_vectors_copy(copy_vector) # copy the user supplied vector
                pdb.gimp_image_insert_vectors(image, new_vector, None, -1)
                new_vector = pdb.gimp_item_transform_rotate(new_vector, math.radians(initial_angle), False, 0, 0)
			
                vector_x, vector_y, vector_r =  x_point + (loop * pf_offset_x), y_point + (loop * pf_offset_y), pf_angle * loop
			
                
                if modifier_options == 0: # No modifier added
                    vector_x, vector_y, vector_r =  x_point + (loop * pf_offset_x), y_point + (loop * pf_offset_y), pf_angle * loop
                    #pdb.gimp_message("pf_angle * loop = " + str(pf_angle * loop) )
				
                elif modifier_options == 1: # modifier applied to x coordinate
                    vector_x, vector_y =  x_point + (loop*(pf_offset_x + pattern_list[loop % len(pattern_list)])*pf_offset_x_mod), y_point + (loop * pf_offset_y)
                elif modifier_options == 2: # modifier applied to y coordinate
                    vector_x, vector_y = x_point + (loop * pf_offset_x), y_point + ((loop*(pf_offset_y + pattern_list[loop % len(pattern_list)]))*pf_offset_y_mod)	
                elif modifier_options == 3: # modifier applied to rotation
                    #vector_r = loop * pf_angle + (((pattern_list[loop % len(pattern_list)])))
                    vector_r = loop * pf_angle + (((pattern_list[loop % len(pattern_list)]))*pf_angle_mod)
                    #vector_r = loop * (pf_angle + pf_angle_mod)
                    #pdb.gimp_message("modifier angle = " + str(vector_r) + " : modifier amount = " + str((((pattern_list[loop % len(pattern_list)]))) ) )
                elif modifier_options == 4: # modifier applied to x and y coordinates
                    vector_x, vector_y = x_point + (loop*(pf_offset_x + pattern_list[loop % len(pattern_list)])*pf_offset_x_mod), y_point + ((loop*(pf_offset_y + pattern_list[loop % len(pattern_list)]))*pf_offset_y_mod)
                elif modifier_options == 5: # modifier applied to rotation and x coordinate
                    vector_x, vector_y =  x_point + (loop*(pf_offset_x + pattern_list[loop % len(pattern_list)])*pf_offset_x_mod), y_point + (loop * pf_offset_y)
                    #vector_r = loop * pf_angle + (((pattern_list[loop % len(pattern_list)])))
                    vector_r = loop * pf_angle + (((pattern_list[loop % len(pattern_list)]))*pf_angle_mod)
                elif modifier_options == 6: # modifier applied to rotation and y coordinate
                    vector_x, vector_y = x_point + (loop * pf_offset_x), y_point + ((loop*(pf_offset_y + pattern_list[loop % len(pattern_list)]))*pf_offset_y_mod)
                    #vector_r = loop * pf_angle + (((pattern_list[loop % len(pattern_list)])))
                    vector_r = loop * pf_angle + (((pattern_list[loop % len(pattern_list)]))*pf_angle_mod)
                elif modifier_options == 7: # modifier applied to all
                    vector_x, vector_y = x_point + (loop*(pf_offset_x + pattern_list[loop % len(pattern_list)])*pf_offset_x_mod), y_point + ((loop*(pf_offset_y + pattern_list[loop % len(pattern_list)]))*pf_offset_y_mod)
                    #vector_r = loop * pf_angle + (((pattern_list[loop % len(pattern_list)])))
                    vector_r = loop * pf_angle + (((pattern_list[loop % len(pattern_list)]))*pf_angle_mod)
                
                stroke_loop = 0
                NumberOfStrokes_in_WorkPath,WorkPath_StrokeIDs_List = pdb.gimp_vectors_get_strokes(new_vector)
                while  stroke_loop<NumberOfStrokes_in_WorkPath:
                    Char_Vector_Stroke = WorkPath_StrokeIDs_List[stroke_loop]
                    pdb.gimp_vectors_stroke_translate(new_vector, Char_Vector_Stroke, vector_x, vector_y)
                    pdb.gimp_vectors_stroke_rotate(new_vector, Char_Vector_Stroke, vector_x, vector_y, vector_r)
                    stroke_loop += 1
                # rename path with x y and rotation info
                new_vector_name = str(vector_x) + "-" + str(vector_y) + "-" + str(vector_r)
                pdb.gimp_item_set_name(new_vector, new_vector_name)
			
                pdb.gimp_context_set_line_width(float(path_width))
				
				
                # colour options
                if colour_option == 0: # foreground
                    path_colour = original_fg_colour # all paths set to fg colour
                elif colour_option == 1: # background
                    path_colour = original_bg_colour # all paths set to bg colour
                elif colour_option == 2: # foreground to background blend
                    path_colour = fn_shade_colour(original_fg_colour, original_bg_colour, num_copies, loop)
                elif colour_option == 3: # background to foreground blend
                    path_colour = fn_shade_colour(original_bg_colour, original_fg_colour, num_copies, loop)
                elif colour_option == 4: # alternate shade colours FG - BG in sequence 
                    if travel_loop % 2 == 1:
                        path_colour = fn_shade_colour(original_fg_colour, original_bg_colour, num_copies, loop)
                    else:
                        path_colour = fn_shade_colour(original_bg_colour, original_fg_colour, num_copies, loop)
                    #pdb.gimp_message(str(loop % 2) )
                elif colour_option == 5: # alternate shade colours BG - FG in sequence 
                    if travel_loop % 2 == 1:
                        path_colour = fn_shade_colour(original_bg_colour, original_fg_colour, num_copies, loop)
                    else:
                        path_colour = fn_shade_colour(original_fg_colour, original_bg_colour, num_copies, loop)
                elif colour_option == 6: # use gradient colours in sequence 
                    path_colour = fn_shade_gradient_colour(colours_of_samples,loop)
                elif colour_option == 7: # use gradient colours in reversed sequence 
                    path_colour = fn_shade_gradient_colour(colours_of_samples,loop)
                elif colour_option == 8: # alternate gradient reversal sequence 
                    if travel_loop % 2 == 1:
                        path_colour = fn_shade_gradient_colour(colours_of_reversed_samples,loop)
                    else:
                        path_colour = fn_shade_gradient_colour(colours_of_samples,loop)
                elif colour_option == 9: # use random gradient colours
                    random.seed()
                    col_loop = random.randrange(0, num_copies)
                    path_colour = fn_shade_gradient_colour(colours_of_samples,col_loop)	
                

                pdb.gimp_context_set_foreground(path_colour)
				
				
                if stroke_method == 1:
                    current_brush = pdb.gimp_context_get_brush() # get current brush
                    use_clipboard_as_brush = False
                    if current_brush == "Clipboard Image":
                        if not use_clipboard_as_brush: # we do not want to use the clipboard image as a brush
                            number_of_brushes, brushes_list = pdb.gimp_brushes_get_list("")
                            if number_of_brushes > 2:
                                pdb.gimp_context_set_brush(brushes_list[2])
						
                if stroke_method == 1: # stroke path with brush
                    pdb.gimp_context_set_brush_size(path_width*1.75) # set brush size
				
                pdb.gimp_context_set_stroke_method(stroke_method)
				
                # stroke the path
                pdb.gimp_drawable_edit_stroke_item(active_layer, new_vector) # stroke the path
                pdb.gimp_displays_flush()
				
				
                # options for dealing with the generated paths
                if path_options == 1: # Merge all paths into one - apart from original user path
                    if loop>1: 
            
                        num_merge_strokes, merge_vector_list = pdb.gimp_vectors_get_strokes(new_vector)
                        merge_vector_id = new_vector
                        stroke_loop = 0
                        while stroke_loop<num_merge_strokes:
                            # loop through paths that make up the new vector
                            merge_vector_stroke = merge_vector_list[stroke_loop]
                            # get information for the most recent path
                            TypeOfStrokeInPart_Vector,NumberOfPointsInMerge_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(merge_vector_id, merge_vector_stroke)
                            merge_ListOfControlPoints = ListOfControlPoints
                            # add new path strokes to the merge path
                            new_vector = pdb.gimp_vectors_stroke_new_from_points(old_vector, 0, NumberOfPointsInMerge_Vector, ListOfControlPoints, IsStrokeClosed)
							
                            stroke_loop += 1
                    		
                    		
                    if loop == 1: old_vector = new_vector # this vector will become the merged path 
                    if loop>=2: 
                        active_vector = pdb.gimp_image_get_active_vectors(image)
                        pdb.gimp_image_remove_vectors(image, active_vector)	# remove latest path after merging
                if path_options == 2: pdb.gimp_image_remove_vectors(image, new_vector)
                if path_options == 3 and loop<num_copies: pdb.gimp_image_remove_vectors(image, new_vector) # leave last path for adding more of the same
				
			
                loop += 1
            if get_out_now:
                travel_loop = num_travel_steps
				
            travel_loop += 1 #############

        pdb.gimp_image_remove_vectors(image, copy_vector)
    else: 
        pdb.gimp_message("To use this option please add a path named walk and try again.")
	
	
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
######################################################################################################################################

    #### Useful Functions
	
######################################################################################################################################

def fn_get_min_max_path(vector_id):
    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(vector_id)
    maxX,minX,maxY,minY,stroke_loop = -3000000,3000000,-3000000,3000000,0
    
    while stroke_loop<NumberOfStrokes_in_Path:
        Char_Vector_Stroke = Char_Vector_StrokeIDs_List[stroke_loop]
        TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(vector_id,Char_Vector_Stroke)
        loop = 0
        while loop<len(ListOfControlPoints):
            x = ListOfControlPoints[loop + 2]
            y = ListOfControlPoints[loop + 3]
            if x<minX: minX = x
            if x>maxX: maxX = x
            if y<minY: minY = y
            if y>maxY: maxY = y
            loop += 6
			
        stroke_loop += 1
    centreX = (maxX - minX)/2
    centreY = (maxY - minY)/2
    return maxX,minX,maxY,minY,centreX,centreY

	
######################################################################################################################################

def fn_shade_gradient_colour(colours_of_samples, instances):
    red_amount = colours_of_samples[(instances*4) + 0] * 255
    green_amount = colours_of_samples[(instances*4) + 1] * 255
    blue_amount = colours_of_samples[(instances*4) + 2] * 255
    alpha_amount = colours_of_samples[(instances*4) + 3] * 255

    if red_amount < 0: red_amount = 0
    if red_amount > 255: red_amount = 255
    if green_amount < 0: green_amount = 0
    if green_amount > 255: green_amount = 255
    if blue_amount < 0: blue_amount = 0
    if blue_amount > 255: blue_amount = 255
    if alpha_amount < 0: alpha_amount = 0
    if alpha_amount > 255: alpha_amount = 255
	
    new_f_colour = (int(red_amount), int(green_amount), int(blue_amount), int(alpha_amount))
    return new_f_colour
	
#######################################################################################################

def fn_get_gradient_positions(instances):
    # get positions along gradients for text colours
    grad_position = 0.0
    if instances<1: instances = 1
    grad_amount = float(1.00/instances)
    loop = 0
    while loop<=instances:
        if loop == 0:
            list_of_pos = [grad_position]
        else:
            list_of_pos.append(grad_position)
        grad_position += grad_amount
        loop += 1
    return list_of_pos

#######################################################################################################
	
def fn_shade_colour(f_colour, bl_colour, instances, amount):
    red_amount_old = f_colour[0]
    green_amount_old = f_colour[1]
    blue_amount_old = f_colour[2]
    red_amount_new = bl_colour[0]
    green_amount_new = bl_colour[1]
    blue_amount_new = bl_colour[2]
    zero = 0.0
	
    if red_amount_old<red_amount_new:
        red_spread = float(red_amount_new - red_amount_old)
        red_step = float(red_spread/instances)
    else:
        red_spread = float(red_amount_old - red_amount_new)
        red_step = float(zero - (red_spread/instances))
		
    if green_amount_old<green_amount_new:
        green_spread = float(green_amount_new - green_amount_old)
        green_step = float(green_spread/instances)
    else:
        green_spread = float(green_amount_old - green_amount_new)
        green_step = float(zero - float((green_spread/instances)))

    if blue_amount_old<blue_amount_new:
        blue_spread = float(blue_amount_new - blue_amount_old)
        blue_step = float(blue_spread/instances)
    else:
        blue_spread = float(blue_amount_old - blue_amount_new)
        blue_step = float(zero - float((blue_spread/instances)))

    new_red_amount = red_amount_old + (red_step * amount)
    if new_red_amount < 0: new_red_amount = 0
    if new_red_amount > 255: new_red_amount = 255
	
    new_green_amount = green_amount_old + (green_step * amount)
    if new_green_amount < 0: new_green_amount = 0
    if new_green_amount > 255: new_green_amount = 255
	
    new_blue_amount = blue_amount_old + (blue_step * amount)
    if new_blue_amount < 0: new_blue_amount = 0
    if new_blue_amount > 255: new_blue_amount = 255
	
    new_f_colour = (int(new_red_amount), int(new_green_amount), int(new_blue_amount))
	
    return new_f_colour
	
######################################################################################################################################
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value

######################################################################################################################################
######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	
######################################################################################################################################
		
def proc_trim_away(img, layer, amount):
    #### Trim off pixels from a layer
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
    #### End of Proc_trim_away
	
######################################################################################################################################
	
######################################################################################################################################
	
register(
    "python_fu_Path_scale_copy_rotate_shift_walk_v_006_Tabbed",
    "Provides options for scaling, rotating, moving and stroking a user path along a walk path.",
    "Provides options for scaling, rotating, moving and stroking a user path along a walk path.",
    "Brian Hahn, Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2011 - 2024",
    "<Image>/Filters/Path Scale Copy Rotate Shift Walk v 0.06 Tabbed...",
    "*",
    [
    (PF_SPINNER, "path_width", "Path width: ", 2, (1, 50, 1)),
    (PF_OPTION, "method", "Stroke path with: ", 0, ("Line", "Currently selected brush")),
    (PF_OPTION, "colour_options", "Colour option: ", 0, ("Use current foreground colour", "Use current background colour", "Shade colours (FG-BG)", "Shade colours (BG-FG)", "Alternate shade (FG-BG)", "Alternate shade (BG-FG)","Use current gradient in sequence","Use current gradient in reversed sequence", "Alternate gradient reversal" , "Use current gradient randomly")),
    (PF_SPINNER, "copies", "Number of times to copy user path: ", 20, (2, 50, 1)),
    (PF_SPINNER, "offset_X", "Stroke x offset (px): ", 0, (-1000, 1000, 1)),
    (PF_SPINNER, "offset_Y", "Stroke y offset (px): ", 0, (-1000, 1000, 1)),
    (PF_SPINNER, "initial_rotation_angle", "Starting angle: ", 0.0, (-360.0, 360.0, 0.1)),
    (PF_SPINNER, "rotation_angle", "Angle of rotation : ", 5.0, (-180.0, 180.0, 0.1)),
    (PF_OPTION, "revolution", "Force revolution amount: ", 0, ("None", "Force one full revolution", "Force one half-revolution", "Force one quarter-revolution", "Force one three quarter-revolution")),
    (PF_SPINNER, "repeats", "Number of repeated patterns along walk path: ", 1, (1, 50, 1)),

    (PF_SPINNER, "scale", "Scale each pattern amount: ", 0.00, (-0.99, 2.00, 0.01),{'newLine':'1','tab':1}),
    (PF_OPTION, "compensate", "Scale compensation positioning: ", 0, ("Off (even spacing)", "On (compensated spacing)"),{'newLine':'1','tab':1}),
    (PF_SPINNER, "mod_offset_X", "Stroke x offset modifier: ", 0.1, (0.01, 0.99, 0.01),{'newLine':'1','tab':1}),
    (PF_SPINNER, "mod_offset_Y", "Stroke y offset modifier: ", 0.1, (0.01, 0.99, 0.01),{'newLine':'1','tab':1}),
    (PF_SPINNER, "mod_rotation", "Angle rotation modifier: ", 0.5, (0.01, 10.00, 0.01),{'newLine':'1','tab':1}),
    (PF_OPTION, "modifier_options", "Pattern modifier options: ", 0, ("No modifiers used","Apply to x offset", "Apply to y offset", "Apply to rotation", "Apply to x and y offsets", "Apply to x offset and rotation ", "Apply to y offset and rotation", "Apply to all"),{'newLine':'1','tab':1}),
    (PF_SPINNER, "path_modifier", "Pattern modifier steps: ", 10.0, (0.0, 50.0, 1.0),{'newLine':'1','tab':1}),
    (PF_OPTION, "modifier_type", "Modifier sequence style: ", 0, ("Ripple", "Ascend jump", "Descend jump", "Zigzag (large)", "Zigzag (small)"),{'newLine':'1','tab':1}),
    (PF_OPTION, "path_options", "Generated path options: ", 2, ("Keep generated paths", "Merge generated paths at each repeat", "Remove all generated paths"),{'newLine':'1','tab':1}),
    (PF_TAB    , "ptab"    , "tabs", 0, (_('Base settings'),_('Scaling options, Modifier options, Path options')) ),
    ],
    [],
    do_stuff)

main()
